#!/usr/bin/env python

"""
Generic file access.

Copyright (C) 2009, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.data import vint, vint_to_array
from array import array
import zlib

# Constants.

class File:

    "A basic file abstraction."

    def __init__(self, f):
        self.f = f
        self.reset()

    def reset(self):

        "To be used to reset the state of the reader or writer between records."

        pass

    def seek(self, offset):
        self.f.seek(offset)
        self.reset()

    def rewind(self):
        self.f.seek(0)
        self.reset()

    def close(self):
        if self.f is not None:
            self.f.close()
            self.f = None

    def get_value_size(self, value):
        if isinstance(value, (list, tuple)):
            return len(value)
        else:
            return 0

    def get_initial_value(self, size):
        if size:
            return [0] * size
        else:
            return 0

class FileWriter(File):

    "Writing basic data types to files."

    def __init__(self, f):
        File.__init__(self, f)

    def write_number(self, number):

        "Write 'number' to the file using a variable length encoding."

        self.f.write(vint(number))

    def write_string(self, s, compress=0):

        """
        Write 's' to the file, recording its length and compressing the string
        if 'compress' is set to a true value.
        """

        # Convert Unicode objects to strings.

        if isinstance(s, unicode):
            s = s.encode("utf-8")

        # Compress the string if requested.

        if compress:
            cs = zlib.compress(s)

            # Take any shorter than the original.

            if len(cs) < len(s):
                flag = "z"
                s = cs
            else:
                flag = "-"

        else:
            flag = ""

        # Write the length of the data before the data itself.

        length = len(s)
        self.f.write("".join([flag, vint(length), s]))

    def write_sequence(self, output, value, last, size, monotonic=1):
        if size:
            emit_delta = 1
            for v, l in map(None, value, last)[:size]:
                if v is None:
                    v = l
                if monotonic or emit_delta:
                    v_out = v - l
                    if emit_delta and v_out != 0:
                        emit_delta = 0
                else:
                    v_out = v + 1
                vint_to_array(v_out, output)
        else:
            vint_to_array(value - last, output)

        return value

class FileReader(File):

    "Reading basic data types from files."

    def __init__(self, f):
        File.__init__(self, f)

    def read_number(self):

        "Read a number from the file."

        # Read each byte, adding it to the number.

        a = array('B')
        fromfile = a.fromfile
        f = self.f

        fromfile(f, 1)
        csd = a[-1]
        if csd < 128:
            return csd
        else:
            while csd & 128:
                fromfile(f, 1)
                csd = a[-1]
            return sum([((csd & 127) << (number * 7)) for (number, csd) in enumerate(a)])

    def read_string(self, decompress=0):

        """
        Read a string from the file, decompressing the stored data if
        'decompress' is set to a true value.
        """

        read = self.f.read

        # Decompress the data if requested.

        if decompress:
            flag = read(1)
        else:
            flag = "-"

        length = self.read_number()
        s = read(length)

        # Perform decompression if applicable.

        if flag == "z":
            s = zlib.decompress(s)

        # Convert strings to Unicode objects.

        return unicode(s, "utf-8")

    def read_sequence(self, last, size, monotonic=1):
        if size:
            expect_delta = 1
            value = []
            for v in last:
                v_in = self.read_number()
                if monotonic or expect_delta:
                    value.append(v + v_in)
                    if expect_delta and v_in != 0:
                        expect_delta = 0
                else:
                    value.append(v_in - 1)
            return tuple(value)
        else:
            return last + self.read_number()

# vim: tabstop=4 expandtab shiftwidth=4
