#!/usr/bin/env python

"""
Specific classes for storing term information.

Copyright (C) 2009, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.files import *
from iixr.positions import *
from iixr.phrases import PhraseIterator
from os.path import commonprefix # to find common string prefixes
from bisect import bisect_right  # to find terms in the dictionary index

class TermWriter(FileWriter):

    "Writing term information to files."

    def reset(self):
        self.last_term = ""
        self.last_offset = 0

    def write_term(self, term, offset, frequency, doc_frequency):

        """
        Write the given 'term', its position file 'offset', its 'frequency' and
        its 'doc_frequency' (number of documents in which it appears) to the
        term information file.
        """

        if term <= self.last_term:
            raise ValueError, "Term %r precedes the previous term %r." % (term, self.last_term)

        # Write the prefix length and term suffix.

        common = len(commonprefix([self.last_term, term]))
        suffix = term[common:]

        self.write_number(common)
        self.write_string(suffix)

        # Write the offset delta.

        self.write_number(offset - self.last_offset)

        # Write the frequency.

        self.write_number(frequency)

        # Write the document frequency.

        self.write_number(doc_frequency)

        self.last_term = term
        self.last_offset = offset

class TermReader(FileReader):

    "Reading term information from files."

    def reset(self):
        self.last_term = ""
        self.last_offset = 0

    def read_term(self):

        """
        Read a term, its position file offset, its frequency and its document
        frequency from the term information file.
        """

        # Read the prefix length and term suffix.

        common = self.read_number()
        suffix = self.read_string()

        self.last_term = self.last_term[:common] + suffix

        # Read the offset delta.

        self.last_offset += self.read_number()

        # Read the frequency.

        frequency = self.read_number()

        # Read the document frequency.

        doc_frequency = self.read_number()

        return self.last_term, self.last_offset, frequency, doc_frequency

    def go_to_term(self, term, offset, info_offset):

        """
        Seek past the entry for 'term' having 'offset' to 'info_offset'. This
        permits the scanning for later terms from the specified term.
        """

        self.seek(info_offset)
        self.last_term = term
        self.last_offset = offset

class TermIndexWriter(TermWriter):

    "Writing term dictionary index details to files."

    def reset(self):
        TermWriter.reset(self)
        self.last_info_offset = 0

    def write_term(self, term, offset, frequency, doc_frequency, info_offset):

        """
        Write the given 'term', its position file 'offset', its 'frequency' and
        its 'doc_frequency' to the term dictionary index file, along with the
        'info_offset' in the term information file.
        """

        TermWriter.write_term(self, term, offset, frequency, doc_frequency)

        # Write the information file offset delta.

        self.write_number(info_offset - self.last_info_offset)
        self.last_info_offset = info_offset

class TermIndexReader(TermReader):

    "Reading term dictionary index details from files."

    def reset(self):
        TermReader.reset(self)
        self.last_info_offset = 0

    def read_term(self):

        """
        Read a term, its position file offset, its frequency, its document
        frequency and a term information file offset from the term dictionary
        index file.
        """

        term, offset, frequency, doc_frequency = TermReader.read_term(self)

        # Read the offset delta.

        self.last_info_offset += self.read_number()

        return term, offset, frequency, doc_frequency, self.last_info_offset

class TermDictionaryWriter:

    "Writing term dictionaries."

    def __init__(self, info_writer, index_writer, position_dict_writer, interval):
        self.info_writer = info_writer
        self.index_writer = index_writer
        self.position_dict_writer = position_dict_writer
        self.interval = interval
        self.entry = 0

    def _write_term(self, term, offset, frequency, doc_frequency):

        """
        Write the given 'term', its position file 'offset', its 'frequency' and
        its 'doc_frequency' (number of documents in which it appears) to the
        term information file. Return the offset after the term information was
        written to the file.
        """

        self.info_writer.write_term(term, offset, frequency, doc_frequency)

        if self.entry % self.interval == 0:
            info_offset = self.info_writer.f.tell()
            self.index_writer.write_term(term, offset, frequency, doc_frequency, info_offset)

        self.entry += 1

    def write_term_positions(self, term, doc_positions):

        """
        Write the given 'term' and the 'doc_positions' recording the documents
        and positions at which the term is found.
        """

        offset, frequency, doc_frequency = self.position_dict_writer.write_term_positions(doc_positions)

        if not frequency or not doc_frequency:
            raise ValueError, "Term %r has no occurrences recorded: %r" % (term, doc_positions)

        self._write_term(term, offset, frequency, doc_frequency)

    def close(self):
        self.info_writer.close()
        self.index_writer.close()
        self.position_dict_writer.close()

class TermDictionaryReader:

    "Reading term dictionaries."

    def __init__(self, info_reader, index_reader, position_dict_reader):
        self.info_reader = info_reader
        self.index_reader = index_reader
        self.position_dict_reader = position_dict_reader

        self.terms = []
        try:
            while 1:
                self.terms.append(self.index_reader.read_term())
        except EOFError:
            pass

        # Large numbers for ordering purposes.

        if self.terms:
            self.max_offset = self.terms[-1][1] + 1
        else:
            self.max_offset = None

    def _find_closest_entry(self, term):

        """
        Find the offsets and frequencies of 'term' from the term dictionary or
        the closest term starting with the value of 'term'.

        Return the closest index entry consisting of a term, the position file
        offset, the term frequency, the document frequency, and the term details
        file offset.
        """

        i = bisect_right(self.terms, (term, self.max_offset, 0, 0)) - 1

        # Get the entry position providing the term or one preceding it.
        # If no entry precedes the requested term, return the very first entry
        # as the closest.

        if i == -1:
            return self.terms[0]
        else:
            return self.terms[i]

    def _find_closest_term(self, term):

        """
        Find the offsets and frequencies of 'term' from the term dictionary or
        the closest term starting with the value of 'term'.

        Return the closest term (or the term itself), the position file offset,
        the term frequency, the document frequency, and the term details file
        offset (or None if the reader is already positioned).
        """

        found_term, offset, frequency, doc_frequency, info_offset = self._find_closest_entry(term)

        # Where the term is found immediately, return the offset and
        # frequencies. If the term does not appear, return the details of the
        # closest entry.

        if term <= found_term:
            return found_term, offset, frequency, doc_frequency, info_offset

        # Otherwise, seek past the index term's entry in the information file
        # and scan for the desired term.

        else:
            self.info_reader.go_to_term(found_term, offset, info_offset)
            try:
                while term > found_term:
                    found_term, offset, frequency, doc_frequency = self.info_reader.read_term()
            except EOFError:
                pass

            return found_term, offset, frequency, doc_frequency, None

    def _find_term(self, term):

        """
        Find the position file offset and frequency of 'term' from the term
        dictionary.
        """

        found_term, offset, frequency, doc_frequency, info_offset = self._find_closest_term(term)

        # If the term is found, return the offset and frequencies.

        if term == found_term:
            return offset, frequency, doc_frequency
        else:
            return None

    def _get_term_and_positions(self, term, offset, frequency, doc_frequency):

        """
        Return the term plus positions details using the given 'term', 'offset',
        'frequency' and 'doc_frequency'.
        """

        return term, frequency, doc_frequency, self._get_positions(offset, doc_frequency)

    def _get_positions(self, offset, doc_frequency):

        """
        Obtain positions from the position index 'offset' expecting a number of
        documents equal to the given 'doc_frequency'.
        """

        return self.position_dict_reader.read_term_positions(offset, doc_frequency)

    # Iterator convenience methods.

    def __iter__(self):
        self.rewind()
        return self

    def next(self):
        try:
            return self.read_term()
        except EOFError:
            raise StopIteration

    # Sequential access methods.

    def rewind(self):
        self.info_reader.rewind()

    def read_term(self):

        """
        Return the next term, its frequency, its document frequency, and the
        documents and positions at which the term is found.
        """

        term, offset, frequency, doc_frequency = self.info_reader.read_term()
        return self._get_term_and_positions(term, offset, frequency, doc_frequency)

    def go_to_term(self, term):

        """
        Navigate to 'term' in the dictionary, returning the details from its
        entry. The returned details can be augmented with position information
        when presented to the _get_term_and_positions method.
        """

        found_term, offset, frequency, doc_frequency, info_offset = self._find_closest_term(term)

        # Position the reader, if necessary.

        if info_offset is not None:
            self.info_reader.go_to_term(found_term, offset, info_offset)

        return found_term, offset, frequency, doc_frequency

    # Query methods.

    def get_terms(self):

        "Return a list of all terms."

        return iter(self)

    def find_terms(self, term):

        "Return all terms whose values start with the value of 'term'."

        terms = []

        found_term, offset, frequency, doc_frequency = self.go_to_term(term)

        # Read and record terms.

        try:
            # Add the found term if it starts with the specified term.

            while found_term.startswith(term):
                terms.append(found_term)
                found_term, offset, frequency, doc_frequency = self.info_reader.read_term()

        except EOFError:
            pass

        return terms

    def find_positions(self, term):

        "Return the documents and positions at which the given 'term' is found."

        t = self._find_term(term)
        if t is None:
            return []
        else:
            offset, frequency, doc_frequency = t
            return self._get_positions(offset, doc_frequency)

    def find_common_positions(self, terms):

        """
        Return the documents and positions at which all the given 'terms' are
        found, where only common documents are returned.
        """

        return PhraseIterator([self.find_positions(term) for term in terms])

    def get_frequency(self, term):

        "Return the frequency of the given 'term'."

        t = self._find_term(term)
        if t is None:
            return None
        else:
            offset, frequency, doc_frequency = t
            return frequency

    def get_document_frequency(self, term):

        "Return the document frequency of the given 'term'."

        t = self._find_term(term)
        if t is None:
            return None
        else:
            offset, frequency, doc_frequency = t
            return doc_frequency

    def close(self):
        self.info_reader.close()
        self.index_reader.close()
        self.position_dict_reader.close()

# vim: tabstop=4 expandtab shiftwidth=4
